/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <wchar.h>

#include <iomanip>

#include <stdlib.h>
#include <getopt.h>

#include "libtkcommon/config.h"
#include "libtkcommon/tkintl.h"
#include "libtkalgorithm/algorithm.h"
#include "libtkrandom/decision_tree.h"
#include "libtkresolver/display.h"

using namespace Tk;
using namespace Random;


void
print_help()
{
  std::cout  << PACKAGE_NAME << " version " << PACKAGE_VERSION << std::endl;
  std::wcout << L"Copyright (C) 2005 Robert Ham (rah@bash.sh)" << std::endl;
  std::wcout << std::endl;
  std::wcout << L"This program comes with ABSOLUTELY NO WARRANTY.  You are licensed to use it" << std::endl;
  std::wcout << L"under the terms of the GNU General Public License, version 3 or later.  See" << std::endl;
  std::wcout << L"the COPYING file that came with this software for details." << std::endl;
  std::wcout << std::endl;
  std::wcout << L" -h, --help                  " << _("Display this help info") << std::endl;
  std::wcout << L" -a, --algorithm             " << _("Display algorithm information") << std::endl;
  std::wcout << L" -b, --bits                  " << _("Bit descriptor") << std::endl;
  std::wcout << L" -s, --speed                 " << _("Display bandwidth requirements") << std::endl;
  std::wcout << L" -c, --traces                " << _("Number of traces") << std::endl;
  std::wcout << std::endl;
}


int
main(int argc, char ** argv)
{
  String * spec(0);
  double speed(0.0);
  unsigned short ntraces(1);
  unsigned bit_desc(1);

  int opt;
  const char * options = "ha:s:c:b:";
  struct option long_options[] = {
    { "help",        no_argument,       NULL, 'h' },
    { "algorithm",   required_argument, NULL, 'a' },
    { "bits",        required_argument, NULL, 'b' },
    { "speed",       required_argument, NULL, 's' },
    { "traces",      required_argument, NULL, 'c' },
    { 0, 0, 0, 0 }
  };
  
  while ((opt = getopt_long (argc, argv, options, long_options, NULL)) != -1) {
    switch (opt) {
      
    case 'a':
      spec = new String(optarg);
      break;

    case 'b':
      bit_desc = strtoul(optarg, NULL, 10);
      break;

    case 's':
      speed = strtod(optarg, NULL);
      break;

    case 'c':
      ntraces = strtoul(optarg, NULL, 10);
      break;

    case 'h':
      print_help ();
      exit (0);
      break;
    case ':':
    case '?':
      print_help ();
      exit (1);
      break;
    }
  }

  if (!spec)
    {
      std::wcerr << _(L"You must provide an algorithm spec")
		 << std::endl
		 << std::endl;
      print_help();
    }




  const Algorithm algorithm(*spec);
  const DecisionTree decision_tree(algorithm.get_outcome_count(), bit_desc);

  const SmallNumbers & offsets(algorithm.get_offsets());
  const SmallNumbers & weights(algorithm.get_weights());
  OffsetProbabilities probs(algorithm.get_probabilities());

  std::wcout << std::endl
    
	     << L'\t'
	     << L'\t'
	     << _(L"Algorithm: ")
	     << bit_desc
	     << L": "
	     << *spec
	     << std::endl
	     << std::endl
	
	     << L'\t'
	     << _(L"Outcome count:  ")
	     << algorithm.get_outcome_count()
	     << std::endl
    
	     << L'\t'
	     << _(L"Tree depth:     ")
	     << decision_tree.get_depth()
	     << std::endl

	     << L'\t'
	     << _(L"Tree outcomes:  ")
	     << decision_tree.tree_outcomes()
	     << std::endl
	
	     << L'\t'
	     << _(L"Expected tries: ")
	     << decision_tree.expected_tries()
	     << std::endl
	
	     << std::endl
    
	     << L'\t'
	     << _(L"Offsets")
	     << std::endl;

  SmallNumbers::const_iterator offset( offsets.begin() );
  SmallNumbers::const_iterator weight( weights.begin() );
  for (; offset != offsets.end(); ++offset, ++weight)
    {
      std::wcout << L"\t["
		 << std::setiosflags(std::ios_base::fmtflags(0))
		 << std::setw(2)
		 << std::right
		 << *offset
		 << L"]:\t"
		 
		 << _(L"weight: ")
		 << std::setw(4)
		 << std::right
		 << *weight
		 << L'\t'

		 << L"p: "
		 << std::fixed
		 << std::setprecision(8)
		 << std::setw(6)
		 << String( probs[*offset] )

		 << std::endl;
    }

  const SmallNumbers & outcomes(algorithm.get_outcome_pattern());
  const unsigned short line_total = 13;
  unsigned outcome_total = 0;

  SmallNumbers::const_iterator outcome( outcomes.begin() );
  std::wcout << std::endl
	     << L'\t'
	     << std::setiosflags(std::ios_base::fmtflags(0))
	     << _(L"Outcome pattern")
	     << std::endl
	     << std::right;

  do
    {
      std::wcout << L'\t';

      for (unsigned short i = 0;
	   i < line_total
	     && outcome_total < algorithm.get_outcome_count();
	   ++i, ++outcome, ++outcome_total)
	{
	  std::wcout << L'['
		     << std::setw(2)
		     << *outcome
		     << L']';

	}
      
      std::wcout << std::endl;
    }
  while (outcome_total < algorithm.get_outcome_count());


  if (speed > 0.0)
    {
      std::wcout << std::endl;
      
      std::wcout << L'\t'
		 << std::setiosflags(std::ios_base::fmtflags(0))
		 << _(L"Bandwidth requirements for ")
		 << ntraces
		 << _(L" traces, ")
		 << std::setprecision(2)
		 << speed
		 << _(L" screens/sec")
		 << std::endl;

      const unsigned screen_heights[5] = {
	1200,
	1024,
	768,
	600,
	480
      };
      
      double bits_per_second;
      for (unsigned short i = 0; i < 5; ++i)
	{
	  bits_per_second = decision_tree.get_depth() * speed * screen_heights[i] * ntraces * bit_desc;

	  std::wcout << L"\tscreen height "
		     << std::setiosflags(std::ios_base::fmtflags(0))
		     << std::setw(4)
		     << std::right
		     << screen_heights[i]
		     << L":\t"
	  
		     << std::setprecision(2)
		     << bits_per_second
		     << L" bps"

		     << L" ("
		     << (bits_per_second * decision_tree.expected_tries())
		     << _(L" expected)")

		     << std::endl;

	}
    }
  
  std::wcout << std::endl;

  exit (0);
}
