/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <boost/variant/apply_visitor.hpp>
#include <odbc++/resultsetmetadata.h>

#include "libtkdb/value.h"
#include "libtkdb/database.h"

using namespace Tk::Db::Data;

Type
Value:: get_type(const Tk::String &property_value_type_name)
{
  if ("boolean" == property_value_type_name)
    return Boolean;
  else if ("integer" == property_value_type_name)
    return Integer;
  else if ("unsigned" == property_value_type_name)
    return Unsigned;
  else if ("double" == property_value_type_name)
    return Double;
  else if ("string" == property_value_type_name)
    return String;
  else if ("timestamp" == property_value_type_name)
    return Timestamp;
  else if ("date" == property_value_type_name)
    return Date;
  else if ("blob" == property_value_type_name)
    return Blob;
  else
    throw UnknownType(property_value_type_name, "database property value");
}

int
Value:: type_to_odbc_type(Type type)
{
       if (Boolean   == type) return odbc::Types::BIT;
  else if (Integer   == type) return odbc::Types::INTEGER;
  else if (Unsigned  == type) return odbc::Types::BIGINT;
  else if (Double    == type) return odbc::Types::DOUBLE;
  else if (String    == type) return odbc::Types::VARCHAR;
  else if (Timestamp == type) return odbc::Types::TIMESTAMP;
  else if (Date      == type) return odbc::Types::DATE;
  else if (Blob      == type) return odbc::Types::LONGVARBINARY;
  else throw std::runtime_error("Unknown data value type (" + static_cast<std::string>(Tk::String(type)) + ")");
}

Type
Value:: odbc_type_to_type(int odbc_type)
{
  switch (odbc_type)
    {
    case odbc::Types::BIT:
    case odbc::Types::TINYINT:
      return Boolean;
    case odbc::Types::SMALLINT:
    case odbc::Types::INTEGER:
      return Integer;
    case odbc::Types::BIGINT:
      return Unsigned;
    case odbc::Types::FLOAT:
    case odbc::Types::DOUBLE:
    case odbc::Types::REAL:
    case odbc::Types::NUMERIC:
    case odbc::Types::DECIMAL:
      return Double;
    case odbc::Types::VARCHAR:
    case odbc::Types::CHAR:
      return String;
    case odbc::Types::TIMESTAMP:
      return Timestamp;
    case odbc::Types::DATE:
      return Date;
    case odbc::Types::LONGVARBINARY:
      return Blob;
    default:
      throw std::runtime_error("Unknown ODBC type (" + static_cast<std::string>(Tk::String(odbc_type)) + ")");
    }
}

Value:: Value(Type type, const Datum &datum)
  : _type(type), _datum(datum)
{
}

Value:: Value(DataSet &result, int index)
{
  odbc::ResultSetMetaData *meta( result->getMetaData() );
  int type = meta->getColumnType(index);
  
  _type = odbc_type_to_type(type);
  
  if (Boolean == _type)
    _datum = result->getBoolean(index);
  else if (Integer == _type)
    _datum = result->getInt(index);
  else if (Unsigned == _type)
    _datum = static_cast<unsigned>(result->getLong(index));
  else if (Double == _type)
    _datum = result->getDouble(index);
  else if (String == _type)
    _datum = static_cast<Tk::String>(result->getString(index));
  else if (Timestamp == _type)
    _datum = result->getTimestamp(index);
  else if (Date == _type)
    _datum = result->getDate(index);
  else if (Blob == _type)
    _datum = Bytes(*result->getBinaryStream(index));
  else
    throw UnknownType(Tk::String(_type), "ODBC-derived Value");
  
  if (result->wasNull())
    _datum.clear();
}

Type Value::type_of(const bool *)             { return Boolean; }
Type Value::type_of(const int *)              { return Integer; }
Type Value::type_of(const unsigned *)         { return Unsigned; }
Type Value::type_of(const double *)           { return Double; }
Type Value::type_of(const Tk::String *)       { return String; }
Type Value::type_of(const Tk::Db::DateTime *) { return Timestamp; }
Type Value::type_of(const Tk::Db::Date *)     { return Date; }
Type Value::type_of(const Tk::Db::Bytes *)    { return Blob; }


Type
Value:: get_type() const
{
  return _type;
}

const Datum &
Value:: get_datum() const
{
  return _datum;
}


class ParamSetter
  : public boost::static_visitor<>
{
private:
  Tk::Db::PStatement *_stmt;
  unsigned            _idx;
public:
  ParamSetter(Tk::Db::PStatement *stmt, unsigned idx) : _stmt(stmt), _idx(idx) {}
  void operator()(const bool &v)               { _stmt->setBoolean(_idx, v); }
  void operator()(const int &v)                { _stmt->setInt(_idx, v); }
  void operator()(const unsigned &v)           { _stmt->setLong(_idx, v); }
  void operator()(const double &v)             { _stmt->setDouble(_idx, v); }
  void operator()(const Tk::String &v)         { _stmt->setString(_idx, v); }
  void operator()(const Tk::Db::DateTime &v)   { _stmt->setTimestamp(_idx, v); }
  void operator()(const Tk::Db::Date &v)       { _stmt->setDate(_idx, v); }
  //void operator()(const Tk::Db::Bytes &v)      { _stmt->setBytes(_idx, v); }
  void operator()(Tk::Db::Bytes &v)            {
    Tk::Stream *strm( v.get_stream() );
    WDEBUG_OUT << L"Writing Bytes with size `"
	       << Tk::String(v.getSize())
	       << L"', stream length `"
	       << Tk::String(strm->get_len())
	       << L"'"
	       << std::endl;
    _stmt->setBinaryStream(_idx, strm->get_stream(), strm->get_len());
  }
};

void
Value:: set_parameter(Tk::Db::PStatement *stmt, unsigned index)
{
  if (_datum.is_null())
    stmt->setNull(index, type_to_odbc_type(_type));
  else
    {
      ParamSetter ps(stmt, index);
      boost::apply_visitor(ps, _datum.get_value());
    }
}

Values:: Values()
{
}

void
Values:: append(const Values &values)
{
  std::list<Value>::insert(end(), values.begin(), values.end());
}

void
Values:: set_parameters(Tk::Db::PStatement *stmt, unsigned start_index)
{
  unsigned index( start_index );
  for(Values::iterator iter( this->begin() );
      iter != this->end();
      ++iter, ++index)
    iter->set_parameter(stmt, index);
}
