/* -*- c++ -*-
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "libtkquery/query.h"

using namespace Tk::Query;

Query::TableAliasNos Query::_table_alias_nos;

Query::TableNo
Query:: get_table_alias_no(const TableAlias &alias)
{
  TableAliasNos::iterator alias_no( _table_alias_nos.find(alias) );
  
  if (alias_no == _table_alias_nos.end())
    {
      TableNo no(1);
      _table_alias_nos.insert( TableAliasNos::value_type(alias, no) );
      
      return no;
    }
  
  return ++(alias_no->second);
}


Query:: Query(Tk::Db::Database *database)
  : _database(database)
{
}

void
Query:: append_tables(const Query &query)
{
  _tables.insert(_tables.end(), query._tables.begin(), query._tables.end());
  _link_clauses.insert(_link_clauses.end(), query._link_clauses.begin(), query._link_clauses.end());
  _link_params.append(query._link_params);
}

TableName
Query:: get_table_abbrev(const TableName &name)
{
  TableName::value_type c(*name.begin());
  TableName abbrev(c);
  
  for (unsigned i(1); i < (name.size() - 1); ++i)
    {
      if (name[i] == L'_')
	abbrev += TableName(name[i+1]);
    }
  
  return abbrev;
}

TableAlias
Query:: get_table_alias(const TableName &name)
{
  TableName abbrev(get_table_abbrev(name));
  return abbrev + TableAlias(get_table_alias_no(abbrev));
}

TableAlias
Query:: add_table(const TableName &name)
{
  TableAlias alias( get_table_alias(name) );
  _tables.push_back(TableDetails(name, alias));
  return alias;
}



TableAlias
Query:: add_property_value_table(const Tk::String &domain, const Tk::String &property_type_name)
{
  Tk::String table_name;
  _database->property_type_get_details(domain, property_type_name, 0, 0, &table_name);
  
  return add_table(table_name);
}

void
Query:: add_link_clause(const Tk::String &clause, const Tk::Db::Data::Values &parameters)
{
  _link_clauses.push_back(clause);
  _link_params.append(parameters);
}

void
Query:: add_link_clause(const Tk::String &alias1,
			const Tk::String &column1,
			const Tk::String &alias2,
			const Tk::String &column2)
{
  Tk::String clause( alias1 );
  clause += Tk::String(".");
  clause += column1;
  clause += Tk::String(" = ");
  clause += alias2;
  clause += Tk::String(".");
  clause += column2;
  
  add_link_clause(clause);
}

void
Query:: set_expr(const Tk::String &expr, const Tk::Db::Data::Values &parameters)
{
  _expr = expr;
  _expr_params = parameters;
}

const Tk::String &
Query:: get_expr() const
{
  return _expr;
}

const Tk::Db::Data::Values &
Query:: get_expr_params() const
{
  return _expr_params;
}


Tk::String
Query:: get_sql(const Tk::String &domain) const
{
  TableAlias domain_alias( get_table_abbrev(domain) );
  Tk::String domain_id( domain + Tk::String("_id") );

  std::wostringstream sqls;
  
  sqls << Tk::String("select ")
       << domain_alias
       << Tk::String(".")
       << domain_id
       << std::endl
       << Tk::String("from   ")
       << domain
       << Tk::String(" ")
       << domain_alias;
  
  for (TableList::const_iterator iter( _tables.begin() );
       iter != _tables.end();
       ++iter)
    {
      sqls << std::endl
	   << Tk::String(",      ")
	   << iter->first
	   << Tk::String(" ")
	   << iter->second;
    }
  
  sqls << std::endl
       << Tk::String("where  ");
  for (Tk::StringList::const_iterator iter( _link_clauses.begin() );
       iter != _link_clauses.end();
       ++iter)
    {
      if (iter != _link_clauses.begin())
	sqls << std::endl
	     << Tk::String("and    ");
      sqls << *iter;
    }
  
  sqls << std::endl
       << Tk::String("and    (")
       << std::endl
       << Tk::String("         ")
       << _expr
       << std::endl
       << Tk::String("       )");
  
  //  WDEBUG_OUT << L"sql: " << sqls.str() << std::endl;
  return sqls.str();
}

Tk::Db::PStatement *
Query:: get_statement(const Tk::String &domain) const
{
  Tk::Db::Data::Values params( _link_params );
  params.append(_expr_params);
  Tk::Db::PStatement *stmt( _database->prepare_statement(get_sql(domain)) );
  params.set_parameters(stmt);
  
  return stmt;
}

Tk::Db::Ids
Query:: get_entities(const Tk::String &domain) const
{
  Tk::Db::PStatementPtr stmt( get_statement(domain) );
  return _database->get_ids(*stmt);
}


Tk::Db::Ids
Tk::Query:: get_all_entities(Db::Database *database, const Tk::String &domain)
{
  std::wostringstream sqls;
  sqls << L" select   d." << domain << "_id"
       << L" from     " << domain << L" d" ;
  
  WDEBUG_OUT << L"preparing statement (" << sqls.str() << ").." << std::endl;
  Tk::Db::PStatement *stmt( database->prepare_statement(sqls.str()) );

  if (!stmt)
    throw std::runtime_error("statement is null");

  WDEBUG_OUT << L"..prepared statement; getting entities.." << std::endl;

  Tk::Db::Ids ids( database->get_ids(*stmt) );

  WDEBUG_OUT << L"..returning entities" << std::endl;
  return ids;
}

