/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <stdexcept>

#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "alea.h"
#include "retriever_registry.h"

using namespace Tk;
using namespace Random;

static const unsigned short USB_VENDOR_ARANEUS      = 0x12d8;
static const unsigned short USB_PRODUCT_ALEA        = 0x0001;
static const unsigned short USB_ENDPOINT_RANDOMNESS = 1;
static const unsigned short TIMEOUT                 = 5000; /* milliseconds */
static const unsigned short BLOCK_SIZE              = 64;

struct usb_dev_handle * Alea::_handle (0);
pthread_mutex_t         Alea::_alea_lock = PTHREAD_MUTEX_INITIALIZER;

Alea:: Alea (unsigned count)
  : Retriever (count)
{
  init_static ();
}

void
Alea:: init_static ()
{
  pthread_mutex_lock (&_alea_lock);

  if (Alea::_handle)
    {
      pthread_mutex_unlock (&_alea_lock);
      return;
    }
  
  usb_init();
  usb_find_busses();
  usb_find_devices();
 
  struct usb_bus *    bus;
  struct usb_device * dev;
  bool found (false);

  for (bus = usb_get_busses ();
       bus;
       bus = bus->next)
    {
      for (dev = bus->devices;
	   dev;
	   dev = dev->next)
	{
	  if (dev->descriptor.idVendor  == USB_VENDOR_ARANEUS &&
	      dev->descriptor.idProduct == USB_PRODUCT_ALEA)
	    {
	      found = true;
	      break;
	    }
	}

      if (found)
	break;
  }

  if (!found)
    {
      pthread_mutex_unlock (&_alea_lock);
      throw std::runtime_error ("could not find Alea device");
    }

  std::string alea_usb_device(std::string(bus->dirname) + "/" + dev->filename);

  struct usb_dev_handle * handle = usb_open(dev);
  if (!handle)
    {
      pthread_mutex_unlock (&_alea_lock);
      throw std::runtime_error (std::string("could not open Alea device '") + alea_usb_device + "'");
    }

  int err;
  err = usb_claim_interface (handle, 0);
  if (err < 0)
    {
      pthread_mutex_unlock (&_alea_lock);
      throw UnixError (std::string("could not claim Alea interface on device '")
		       + alea_usb_device + "' (return value: " + static_cast<std::string>(String(err)) + ")", -err);
    }

  /*
  std::wcout << L"Claimed Alea interface on bus "
	     << String(bus->dirname)
	     << L", device "
	     << String(dev->filename)
	     << std::endl;
  */

  _handle = handle;
  pthread_mutex_unlock (&_alea_lock);
}


Buffer *
Alea:: retrieve ()
{
  Buffer * buf = new Buffer;
  char raw_buf[BLOCK_SIZE];
  int nread;

  while (get_count() > 0)
    {
      pthread_mutex_lock (&_alea_lock);
      nread = usb_bulk_read (_handle, USB_ENDPOINT_RANDOMNESS, raw_buf, BLOCK_SIZE, TIMEOUT);
      pthread_mutex_unlock (&_alea_lock);
      
      if (nread < 0)
	throw UnixError ("error reading from Alea endpoint", -nread);

      if (nread < BLOCK_SIZE)
	throw std::runtime_error ("short read from Alea device");

      for (unsigned i = 0;
	   i < BLOCK_SIZE && get_count() > 0;
	   ++i, --(get_count()))
	{
	  buf->push_back (raw_buf[i]);
	}
    }

  return buf;
}

bool
AleaType:: is_true_randomness()
{
  return true;
}

Retriever *
AleaType:: new_retriever(unsigned count)
{
  return new Alea(count);
}

unsigned
AleaType:: get_preferred_block_size()
{
  return BLOCK_SIZE;
}
